#!/usr/bin/python3

## Install
# apt-get install python3-module-pandas python3-module-ldap3

from ldap3 import Server, Connection, ALL
import pandas
from argparse import ArgumentParser
from getpass import getpass
from enum import Enum
import struct

def convert(binary):
    version = struct.unpack('B', binary[0])[0]
    # I do not know how to treat version != 1 (it does not exist yet)
    assert version == 1, version
    length = struct.unpack('B', binary[1])[0]
    authority = struct.unpack('>Q', '\x00\x00' + binary[2:8])[0]
    string = 'S-%d-%d' % (version, authority)
    binary = binary[8:]
    assert len(binary) == 4 * length
    for i in xrange(length):
        value = struct.unpack('<L', binary[4*i:4*(i+1)])[0]
        string += '-%d' % value
    return string

class sAMAccountType(Enum):
    SAM_DOMAIN_OBJECT = 0x0
    SAM_GROUP_OBJECT = 0x10000000
    SAM_NON_SECURITY_GROUP_OBJECT = 0x10000001
    SAM_ALIAS_OBJECT = 0x20000000
    SAM_NON_SECURITY_ALIAS_OBJECT = 0x20000001
    SAM_USER_OBJECT = 0x30000000
    SAM_MACHINE_ACCOUNT = 0x30000001
    SAM_TRUST_ACCOUNT = 0x30000002
    SAM_APP_BASIC_GROUP = 0x40000000
    SAM_APP_QUERY_GROUP = 0x40000001
    SAM_ACCOUNT_TYPE_MAX = 0x7fffffff

parser = ArgumentParser()
parser.add_argument("-a", "--address", type=str, required=True,
                        help="Microsoft AD Domain controller IP Address")
parser.add_argument("-d", "--domain", type=str, required=True,
                        help="Microsoft AD Domain name. Example: dc=domain,dc=alt")
parser.add_argument("-D", "--binddn", type=str, required=True,
                        help="Adminitrative account. Example: Administrator@domain.alt")
parser.add_argument("-w", "--password", type=str, required=False,
                        help="Adminitrative account password")
args = parser.parse_args()

address = args.address
domain = args.domain
binddn = args.binddn

if args.password is None:
    password = getpass()
else:
    password = args.password

def getConnection(address, domain, binddn, password):
    server = Server(address, get_info=ALL)
    conn = Connection(server, binddn, password, auto_bind=True)
    return conn

def countObjects(conn, domain, objectClass, filters = None):
    objects = getObjects(conn, domain, objectClass, filters)
    return len(objects)

def getObjects(conn, domain, objectClass, filters = None, attrs = None):
    request = ''
    if filters == None:
        request = f'(&(objectClass={objectClass}))'
    else:
        request = f'(&(objectClass={objectClass})'
        for filter in filters:
            request += f'({filter}={filters[filter]})'
        request += ')'

    if attrs == None:
        entries = conn.extend.standard.paged_search(domain,
                    request, generator=False)
    else:
        entries = conn.extend.standard.paged_search(domain,
                    request, attributes = attrs, generator=False)

    c = 0
    for entry in entries.copy():
        if 'searchResRef' == entry['type']:
            entries.remove(entry)
        if c <= 10:
            c = c + 1
        else:
            break
    return entries

def countObjectsWithAttrs(conn, domain, objectClass):
    print('')
    print(f'Discover attributes for {objectClass} object')
    currentObjectClass = objectClass;
    attrs = list()
    i = 0
    while i < 30:
        i = i + 1
        conn.search(f'CN=Schema,CN=Configuration,{domain}',
                        f'(&(lDAPDisplayName={currentObjectClass}))',
                        attributes=['mayContain', 'systemMayContain','systemMustContain',
                            'mustContain', 'lDAPDisplayName', 'subClassOf'])
        for entry in conn.entries:
            if 'systemMayContain' in entry:
                attrs.extend([*entry['systemMayContain']])
            if 'mayContain' in entry:
                attrs.extend([*entry['mayContain']])
            if 'systemMustContain' in entry:
                attrs.extend([*entry['systemMustContain']])
            if 'mustContain' in entry:
                attrs.extend([*entry['mustContain']])
 
        if len(conn.entries) > 0:
            print(f'{conn.entries[0]["lDAPDisplayName"]}', end = '')
            if conn.entries[0]['lDAPDisplayName'] == 'top':
                print('')
                break
            else:
                print(' -> ', end = '')
                currentObjectClass = conn.entries[0]['subClassOf']
    print(f'Discover attributes for {objectClass} object complete')

    print(f'Count {objectClass} objects with discovered attributes')
    attrs = list(set(attrs))
    attrs_stat = dict()

    for attr in attrs:
        entries = conn.extend.standard.paged_search(domain,
                    f'(&(objectClass={objectClass})({attr}=*))', generator=False)
        c = 0
        for entry in entries.copy():
            if 'searchResRef' == entry['type']:
                entries.remove(entry)
            if c <= 10:
                c = c + 1
            else:
                break
        attrs_stat[attr] = len(entries)
    print(f'Count {objectClass} objects with discovered attributes complete')

    return attrs_stat

def getCountObjectsByGroups(conn, domain):
    groups = getObjects(conn, domain, 'group', attrs = ['sAMAccountType', 'distinguishedName', 'objectSid', 'member'])
    formattedGroups = dict()
    formattedGroups['sAMAccountType'] = list()
    formattedGroups['BuiltIn'] = list()
    formattedGroups['Group ID'] = list()
    formattedGroups['Members'] = list()
    for group in groups:
        sAMAccountTypeValue = sAMAccountType(group['attributes']['sAMAccountType'])
        formattedGroups['sAMAccountType'].append(sAMAccountTypeValue.name)
        if ''.join(group['attributes']['distinguishedName']).find('Builtin') == -1:
            builtIn = False
        else:
            builtIn = True
        formattedGroups['BuiltIn'].append(builtIn)
        sid = group['attributes']['objectSid'].rpartition('-')[2]
        members = len(group['attributes']['member'])
        users = getObjects(conn, domain, 'user', filters = {'primaryGroupID': sid})
        members += len(users)
        formattedGroups['Group ID'].append(sid)
        formattedGroups['Members'].append(members)

    return formattedGroups

def getCountSubnetsAndControllersBySite(conn, domain):
    sites = getObjects(conn, f'CN=Sites,CN=Configuration,{domain}', 'site', attrs = ['siteObjectBL', 'name'])
    formattedSites = dict()
    formattedSites['Subnets'] = list()
    formattedSites['Domain controllers'] = list()
    for site in sites:
        formattedSites['Subnets'].append(len(site['attributes']['siteObjectBL']))
        siteName = ''.join(site['attributes']['name'])
        servers = getObjects(conn, f'CN=Servers,CN={siteName},CN=Sites,CN=Configuration,{domain}', 'server')
        formattedSites['Domain controllers'].append(len(servers))

    return formattedSites

def getDomainControllersOSVersion(conn, domain):
    controllers = getObjects(conn, f'OU=Domain Controllers,{domain}', 'computer', attrs = ['operatingSystem', 'operatingSystemServicePack', 'operatingSystemVersion'])
    formattedControllers = dict()
    formattedControllers['OS'] = list()
    formattedControllers['OS Service Pack'] = list()
    formattedControllers['OS Version'] = list()
    for controller in controllers:
        operatingSystem = ''.join(controller['attributes']['operatingSystem'])
        formattedControllers['OS'].append(operatingSystem)
        operatingSystemServicePack = ''.join(controller['attributes']['operatingSystemServicePack'])
        formattedControllers['OS Service Pack'].append(operatingSystemServicePack)
        operatingSystemVersion = ''.join(controller['attributes']['operatingSystemVersion'])
        formattedControllers['OS Version'].append(operatingSystemVersion)

    return formattedControllers

def printHeader(name):
    print('====== ' + name)
    print('=========================================================================')

def printFooter():
    print('=========================================================================')
    print('')

def printStat(name, stat):
    printHeader(name)
    pandas.set_option("display.max_rows", None, "display.max_columns", None)
    data = {'Attribute': stat.keys(), 'Number of objects': stat.values()}
    formattedData = pandas.DataFrame(data)
    final_formattedData = formattedData.sort_values('Attribute').reset_index(drop = True)
    print(final_formattedData)
    printFooter()

def printFormatted(name, data):
    printHeader(name)
    pandas.set_option("display.max_rows", None, "display.max_columns", None)
    formattedData = pandas.DataFrame(data)
    final_formattedData = formattedData.sort_values(list(data.keys())[0]).reset_index(drop = True)
    print(final_formattedData)
    printFooter()
    
conn = getConnection(address, domain, binddn, password)

objects = dict()
objects['Users'] = countObjects(conn, domain, 'user')
objects['Groups'] = countObjects(conn, domain, 'group')
objects['Organizational Units'] = countObjects(conn, domain, 'organizationalUnit')
objects['Computers'] = countObjects(conn, domain, 'computer')
objects['Group Policy Containers'] = countObjects(conn, domain, 'groupPolicyContainer')
objects['Containers'] = countObjects(conn, domain, 'container')
objects['Domain Controllers'] = countObjects(conn, domain, 'computer', {'primaryGroupID': '516'})

printStat('Objects count', objects)

objectsByGroups = getCountObjectsByGroups(conn, domain)
printFormatted('Groups summary', objectsByGroups)

subnetsAndControllersBySite = getCountSubnetsAndControllersBySite(conn, domain)
printFormatted('Sites summary', subnetsAndControllersBySite)

domainControllersOSVersion = getDomainControllersOSVersion(conn, domain)
printFormatted('Domain controllers OS information', domainControllersOSVersion)

users_stat = countObjectsWithAttrs(conn, domain, 'User')
groups_stat = countObjectsWithAttrs(conn, domain, 'Group')
ou_stat = countObjectsWithAttrs(conn, domain, 'organizationalUnit')
computers_stat = countObjectsWithAttrs(conn, domain, 'Computer')
gp_containers_stat = countObjectsWithAttrs(conn, domain, 'groupPolicyContainer')
containers_stat = countObjectsWithAttrs(conn, domain, 'Container')

print('')

printStat('User objects', users_stat)
printStat('Group objects', groups_stat)
printStat('Organizational Unit objects', ou_stat)
printStat('Computer objects', computers_stat)
printStat('Group Policy Container objects', gp_containers_stat)
printStat('Container objects', containers_stat)
